// feed.cpp

#include "feed.h"
#include <gdk/gdk.h>
#include <stdlib.h>
#include <string.h>
#include <algorithm>
#include <fstream>
#include <iostream>

// in minutes
#define REFRESH_INTERVAL 30

// utilities

static std::string prefix_homedir (const char *dir)
{
	std::string str (getenv ("HOME"));
	return str + "/" + dir;
}

template <typename T>
static typename std::list<T>::iterator iteratorJump (int nb, typename std::list<T>::iterator begin, typename std::list<T>::iterator end)
{
	typename std::list<T>::iterator it;
	int n = 0;
	for (it = begin; it != end && n < nb; it++) n++;
	return it;
}

template <typename T>
static int iteratorPos (typename std::list<T>::iterator it, typename std::list<T>::iterator begin)
{
	int pos = 0;
	for (typename std::list<T>::iterator i = begin; i != it; i++) pos++;
	return pos;
}

void replace (std::string &str, char dead, char live)
{
	// if we ever need to accept more characters make it accept a pair like
	// "13579" and "24680" to map '1' to '2' and so on.
	for (unsigned int i = 0; i < str.size(); i++)
		if (str[i] == dead)
			str[i] = live;
}

// News

News::News (Feed *feed)
: feed (feed), is_read (false)
{
}

void News::setRead (bool read)
{
	if (is_read != read) {
		is_read = read;
		feed->newsStatusChanged (this);
		if (read)
			feed->read_news.push_back (id);
		else
			feed->read_news.remove (id);
	}
}

void News::setTitle (const std::string &str)
{ _title = str; }
void News::setSummary (const std::string &str)  // content overloads summary
{ if (_summary.empty()) _summary = str; }
void News::setContent (const std::string &str)
{ _summary = str; }
void News::setLink (const std::string &str)
{ _link = str; }
void News::setDate (const std::string &str)
{ _date = str; }
void News::setAuthor (const std::string &str)
{ _author = str; }
void News::addCategory (const std::string &str)
{ if (!_categories.empty()) _categories += ", "; _categories += str; }
void News::setId (const std::string &str)
{
	id = str;  // check if already read
	if (std::find (feed->read_news.begin(), feed->read_news.end(), id) != feed->read_news.end())
		is_read = true;
}

// Feed

Feed::Feed (const std::string &_url, const std::string &title,
            const std::string &codeset)
: url (_url), _title (title), codeset (codeset), _loading (false)
{
}

Feed::~Feed()
{
	clear();
}

void Feed::clear()
{
	for (std::list <News *>::iterator it = news.begin(); it != news.end(); it++)
		delete *it;
	news.clear();
	error_msg.clear();
}

News *Feed::getNews (int nb)
{
	std::list <News *>::iterator it;
	it = iteratorJump<News*> (nb, news.begin(), news.end());
	if (it != news.end()) return *it;
	return NULL;
}

int Feed::getNewsNb (News *n)
{
	std::list <News *>::iterator it = find (news.begin(), news.end(), n);
	return iteratorPos<News*> (it, news.begin());
}

void Feed::parse_done_cb (ParseFeedHandler *handler, const std::string &error)
{
	Feed *pThis = (Feed *) handler;
	pThis->_loading = false;
	pThis->error_msg = error;
	if (error.empty()) {
		// we don't want to keep stored the washed up old flags
		pThis->read_news.clear();
		for (std::list <News *>::const_iterator it = pThis->news.begin();
		     it != pThis->news.end(); it++) {
			if ((*it)->isRead())
				pThis->read_news.push_back ((*it)->id);
		}
	}

	gdk_threads_enter();
	Manager::get()->feedLoaded (pThis);
	gdk_flush();
	gdk_threads_leave();
}

void Feed::refresh()
{
	if (!_loading) {
		_loading = true;
		error_msg.clear();
		Manager::get()->feedLoading (this);
		clear();
		parse (this, url, codeset, parse_done_cb);
	}
}

void Feed::newsStatusChanged (News *news)
{
	Manager::get()->feedStatusChanged (this);
}

int Feed::unreadNb() const
{
	int unread = 0;
	for (std::list <News *>::const_iterator it = news.begin(); it != news.end(); it++)
		if (!(*it)->isRead())
			unread++;
	return unread;
}

void Feed::setUserTitle (const std::string &str)
{ _title = str; }

void Feed::setTitle (const std::string &str)
{ if (_title.empty()) _title = str; _oriTitle = str; }
void Feed::setDescription (const std::string &str)
{ _description = str; }
void Feed::setLink (const std::string &str)
{ _link = str; }
void Feed::setAuthor (const std::string &str)
{ _author = str; }

ParseNewsHandler *Feed::appendNews()
{
	News *n = new News (this);
	news.push_back (n);
	return n;
}

XmlParser::Handler *Feed::startElement (const char *name,
	const char **attribute_names, const char **attribute_values,
	std::string &error)
{
	if (!strcmp (name, "news")) {
		const char *id = XmlParser::get_value ("id", attribute_names, attribute_values);
		if (id)
			read_news.push_back (id);
	}
	return NULL;
}

void Feed::saveConfig (std::ofstream &stream) const
{
	std::string _url (url);  // what the HELL?
	replace (_url, '&', '@');  // check Manager::startElement() for info

	stream << "\t<feed title=\"" << _title << "\" url=\"" << _url << "\"";
	if (!codeset.empty())
		stream << " codeset=\"" << codeset << "\"";
	stream << ">\n";
	// if news empty, then the site is down or the network is; save previous info
	for (std::list <std::string>::const_iterator it = read_news.begin();
	     it != read_news.end(); it++)
		stream << "\t\t<news id=\"" << *it << "\"></news>\n";
	stream << "\t</feed>\n";
}

// Manager

Manager::Manager()
: feeds_loading (0), feeds_loaded (0)
{
	loadConfig();
	g_timeout_add_seconds_full (G_PRIORITY_LOW, REFRESH_INTERVAL*60,
	                            refresh_timeout, this, NULL);
}

Manager *Manager::get()
{
	static Manager *singleton = 0;
	if (!singleton) singleton = new Manager();
	return singleton;
}

gboolean Manager::refresh_timeout (gpointer data)
{
	Manager *pThis = (Manager *) data;
	pThis->saveConfig();  // save sometimes to subsidize a crash
	pThis->refreshAll();
	return TRUE;  // keep going
}

Feed *Manager::addFeed (const std::string &url, const std::string &title,
                        const std::string &codeset)
{
	std::string _url (url);
	if (!_url.compare (0, 7, "feed://")) {
		_url[0] = 'h'; _url[1] = 't'; _url[2] = 't'; _url[3] = 'p';
	}

	notifyStartStructuralChange();
	Feed *feed = new Feed (_url, title, codeset);
	feeds.push_back (feed);
	notifyEndStructuralChange();
	return feed;
}

void Manager::removeFeed (Feed *feed)
{
	notifyStartStructuralChange();
	std::list <Feed *>::iterator it = std::find (feeds.begin(), feeds.end(), feed);
	if (it != feeds.end()) {
		feeds.erase (it);
		delete feed;
	}
	notifyEndStructuralChange();
}

#if 0
void Manager::swap (Feed *feed1, Feed *feed2)
{
	notifyStartStructuralChange();
	std::list <Feed *>::iterator it1 = std::find (feeds.begin(), feeds.end(), feed1);
	std::list <Feed *>::iterator it2 = std::find (feeds.begin(), feeds.end(), feed2);
	(*it1) = feed2;
	(*it2) = feed1;
	notifyEndStructuralChange();
}
#endif

void Manager::move (Feed *feed, int pos)
{
	notifyStartStructuralChange();
	feeds.remove (feed);
	std::list <Feed *>::iterator it;
	it = iteratorJump<Feed*> (pos, feeds.begin(), feeds.end());
	feeds.insert (it, feed);
	notifyEndStructuralChange();
}

void Manager::notifyStartStructuralChange()
{
	for (std::list <Listener *>::iterator it = listeners.begin(); it != listeners.end(); it++)
		(*it)->startStructuralChange (this);
}

void Manager::notifyEndStructuralChange()
{
	for (std::list <Listener *>::iterator it = listeners.begin(); it != listeners.end(); it++)
		(*it)->endStructuralChange (this);
}

Feed *Manager::getFeed (int nb)
{
	std::list <Feed *>::iterator it;
	it = iteratorJump<Feed*> (nb, feeds.begin(), feeds.end());
	if (it != feeds.end()) return *it;
	return NULL;
}

int Manager::getFeedNb (Feed *feed)
{
	std::list <Feed *>::iterator it = find (feeds.begin(), feeds.end(), feed);
	return iteratorPos<Feed*> (it, feeds.begin());
}

void Manager::refreshAll()
{
	for (std::list <Feed *>::iterator it = feeds.begin(); it != feeds.end(); it++)
		(*it)->refresh();
}

int Manager::unreadNb() const
{
	int unread = 0;
	for (std::list <Feed *>::const_iterator it = feeds.begin(); it != feeds.end(); it++)
		unread += (*it)->unreadNb();
	return unread;
}

void Manager::feedStatusChanged (Feed *feed)
{
	for (std::list <Listener *>::iterator it = listeners.begin(); it != listeners.end(); it++)
		(*it)->feedStatusChange (this, feed);
}

void Manager::feedLoading (Feed *feed)
{
	feeds_loading++;
	for (std::list <Listener *>::iterator it = listeners.begin(); it != listeners.end(); it++) {
		(*it)->feedStatusChange (this, feed);
		(*it)->feedLoading (this, feed);
		if (feeds_loading > 1)
			(*it)->feedsLoadingProgress (this, ((float) feeds_loaded) / feeds_loading);
	}
}

void Manager::feedLoaded (Feed *feed)
{
	feeds_loaded++;
	for (std::list <Listener *>::iterator it = listeners.begin(); it != listeners.end(); it++) {
		(*it)->feedStatusChange (this, feed);
		(*it)->feedLoaded (this, feed);
		if (feeds_loading > 1)
			(*it)->feedsLoadingProgress (this, ((float) feeds_loaded) / feeds_loading);
	}
	if (feeds_loaded >= feeds_loading)
		feeds_loading = (feeds_loaded = 0);
}

void Manager::loadConfig()
{
	std::ifstream stream (prefix_homedir (".eatfeed").c_str());
	if (stream.good()) {
		XmlParser parser (this);
		std::string error;
		char buffer [2048];
		while (stream.getline (buffer, 2048)) {
			if (!parser.parse (buffer, error)) {
				std::cout << "Error: couldn't parse .eatfeed: " << error << std::endl;
				break;
			}
		}
	}
	else
		std::cout << "Error: couldn't open .eatfeed for reading.\n";
	atexit (saveManager);
}

XmlParser::Handler *Manager::startElement (const char *name,
	const char **attribute_names, const char **attribute_values,
	std::string &error)
{
	if (!strcmp (name, "feed")) {
		const char *title = "", *url = 0, *codeset = "";
		for (int i = 0; attribute_names[i]; i++) {
			if (!strcmp (attribute_names[i], "title"))
				title = attribute_values[i];
			else if (!strcmp (attribute_names[i], "url"))
				url = attribute_values[i];
			else if (!strcmp (attribute_names[i], "codeset"))
				codeset = attribute_values[i];
		}
		if (url) {
			// gtk xml parser has some adversity to chars on attributes like &
			// I thought of escaping the url at the Feed constrctor ('&' -> '%26')
			// but it seem as sites pass the url to scripts, some don't handle
			// that http escape code
			std::string _url (url);
			replace (_url, '@', '&');

			Feed *feed = addFeed (_url, title, codeset);
			return feed;
		}
	}
	return NULL;
}

void Manager::saveConfig() const
{
	std::ofstream stream (prefix_homedir (".eatfeed").c_str());
	if (stream.good()) {
		stream << "<eatfeed>\n";
		for (std::list <Feed *>::const_iterator it = feeds.begin(); it != feeds.end(); it++)
			(*it)->saveConfig (stream);
		stream << "</eatfeed>\n";
	}
	else
		std::cout << "Error: couldn't open .eatfeed for saving.\n";
}

void Manager::saveManager()
{ Manager::get()->saveConfig(); }

